// XKViewport.c

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/neXtaw/XawInit.h>
#include <X11/Xmu/Misc.h>
#include <XKScrollbar.h>
#include <XKViewportP.h>
#include <X11/neXtaw/ThreeD.h>

static	void	ScrollUpDownProc();
static	void	ThumbProc();
static	Boolean	GetGeometry();
static	void	ComputeWithForceBars();

#define offset(field) XtOffsetOf(XKViewportRec, xkviewport.field)

static	XtResource	resources[] =
{
    { XtNforceBars, XtCBoolean, XtRBoolean, sizeof(Boolean),
		offset(forcebars), XtRImmediate, (XtPointer) False },
    { XtNallowHoriz, XtCBoolean, XtRBoolean, sizeof(Boolean),
		offset(allowhoriz), XtRImmediate, (XtPointer) False },
    { XtNallowVert, XtCBoolean, XtRBoolean, sizeof(Boolean),
		offset(allowvert), XtRImmediate, (XtPointer) False },
    { XtNuseBottom, XtCBoolean, XtRBoolean, sizeof(Boolean),
		offset(usebottom), XtRImmediate, (XtPointer) False },
    { XtNuseRight, XtCBoolean, XtRBoolean, sizeof(Boolean),
		offset(useright), XtRImmediate, (XtPointer) False },
    { XtNreportCallback, XtCReportCallback, XtRCallback, sizeof(XtPointer),
		offset(report_callbacks), XtRImmediate, (XtPointer) NULL },
};

#undef offset

static	void	Initialize();
static	void	ConstraintInitialize();
static	void	Realize();
static	void	Resize();
static	void	ChangeManaged();
static	Boolean	SetValues();
static	Boolean	Layout();

static	XtGeometryResult	GeometryManager();
static	XtGeometryResult	PreferredGeometry();

#define superclass	(&formClassRec)

XKViewportClassRec xkViewportClassRec =
{
	{ // core_class fields
		/* superclass	  		*/	(WidgetClass) superclass,
		/* class_name	  		*/	"XKViewport",
		/* widget_size	  		*/	sizeof(XKViewportRec),
		/* class_initialize		*/	XawInitializeWidgetSet,
		/* class_part_init  	*/  NULL,
		/* class_inited	  		*/	FALSE,
		/* initialize	  		*/	Initialize,
		/* initialize_hook  	*/  NULL,
		/* realize		  		*/	Realize,
		/* actions		  		*/	NULL,
		/* num_actions	  		*/	0,
		/* resources	  		*/	resources,
		/* num_resources	  	*/	XtNumber(resources),
		/* xrm_class	  		*/	NULLQUARK,
		/* compress_motion	  	*/	TRUE,
		/* compress_exposure  	*/	TRUE,
		/* compress_enterleave	*/  TRUE,
		/* visible_interest	  	*/	FALSE,
		/* destroy		  		*/	NULL,
		/* resize		  		*/	Resize,
		/* expose		  		*/	XtInheritExpose,
		/* set_values	  		*/	SetValues,
		/* set_values_hook    	*/  NULL,
		/* set_values_almost  	*/  XtInheritSetValuesAlmost,
		/* get_values_hook    	*/	NULL,
		/* accept_focus	  		*/	NULL,
		/* version            	*/  XtVersion,
		/* callback_private	  	*/	NULL,
		/* tm_table    	  		*/	NULL,
		/* query_geometry     	*/	PreferredGeometry,
		/* display_accelerator	*/	XtInheritDisplayAccelerator,
		/* extension          	*/	NULL
	},
	{ // composite_class fields
		/* geometry_manager	  	*/	GeometryManager,
		/* change_managed	  	*/	ChangeManaged,
		/* insert_child	  		*/	XtInheritInsertChild,
		/* delete_child	  		*/	XtInheritDeleteChild,
		/* extension          	*/	NULL
	},
	{ // constraint_class fields
		/* subresourses	  		*/	NULL,
		/* subresource_count  	*/	0,
		/* constraint_size	  	*/	sizeof(XKViewportConstraintsRec),
		/* initialize	  		*/	ConstraintInitialize,
		/* destroy		  		*/	NULL,
		/* set_values	  		*/	NULL,
		/* extension          	*/	NULL
	},
	{ // form_class fields
		/* layout		  		*/	Layout
	},
	{ // xkviewport_class fields 
		/* empty		  		*/	0
	}
};

WidgetClass xkViewportWidgetClass = (WidgetClass) &xkViewportClassRec;

static Widget CreateScrollbar(w, horizontal)
    XKViewportWidget w;
    Boolean horizontal;
{
    Widget					clip = w->xkviewport.clip;
    XKViewportConstraints	constraints =
		(XKViewportConstraints) clip->core.constraints;

    static	Arg	barArgs[] =
	{
		{ XtNorientation,       (XtArgVal) 0 },
		{ XtNlength,            (XtArgVal) 0 },
		{ XtNleft,              (XtArgVal) 0 },
		{ XtNright,             (XtArgVal) 0 },
		{ XtNtop,               (XtArgVal) 0 },
		{ XtNbottom,            (XtArgVal) 0 },
		{ XtNmappedWhenManaged, (XtArgVal) False },
		{ XtNalwaysVisible,		(XtArgVal) False },
		{ XtNdrawHalf,			(XtArgVal) True },
    };
    Widget				bar;
    Dimension			sw = 0;

    XtVaGetValues((Widget) w->xkviewport.threeD, XtNshadowWidth, &sw, NULL);

    XtSetArg(barArgs[0], XtNorientation,
	      (horizontal) ? XtorientHorizontal : XtorientVertical);
    XtSetArg(barArgs[1], XtNlength,
		 // (horizontal) ? clip->core.width - 2 * sw : clip->core.height - 2 * sw);
	     (horizontal) ? clip->core.width - 2 * sw : w->core.height - 2 * sw);
    XtSetArg(barArgs[2], XtNleft, (!horizontal && w->xkviewport.useright)
		 	? XtChainRight - sw : XtChainLeft + sw);
    XtSetArg(barArgs[3], XtNright, (!horizontal && !w->xkviewport.useright)
		? XtChainLeft + sw : XtChainRight - sw);
    XtSetArg(barArgs[4], XtNtop, (horizontal && w->xkviewport.usebottom)
		? XtChainBottom - sw : XtChainTop + sw);
    XtSetArg(barArgs[5], XtNbottom, (horizontal && !w->xkviewport.usebottom)
		? XtChainTop + sw : XtChainBottom - sw);

    bar = XtCreateWidget((horizontal ? "horizontal" : "vertical"),
		xkScrollbarWidgetClass,
		(Widget) w,
		barArgs, XtNumber(barArgs));
    XtAddCallback(bar, XtNscrollProc, ScrollUpDownProc, (XtPointer) w);
    XtAddCallback(bar, XtNjumpProc, ThumbProc, (XtPointer) w);

    if ((horizontal))
	{
		w->xkviewport.horiz_bar = bar;
		constraints->form.vert_base = bar;
    }
    else
	{
		w->xkviewport.vert_bar = bar;
		constraints->form.horiz_base = bar;
    }
    XtManageChild(bar);
    return bar;
}

/* ARGSUSED */
static void Initialize(request, new, args, num_args)
    Widget request, new;
    ArgList args;
    Cardinal *num_args;
{
    XKViewportWidget	w = (XKViewportWidget) new;
    static	Arg			clip_args[8];
    static	Arg			threeD_args[8];
    Cardinal			arg_cnt;
    Widget				h_bar, v_bar;
    Dimension			clip_height, clip_width;
    Dimension			sw = 0;

    w->form.default_spacing = 0;  // Reset the default spacing to 0 pixels.

	// Initialize all widget pointers to NULL.
    w->xkviewport.child = (Widget) NULL;
    w->xkviewport.horiz_bar = w->xkviewport.vert_bar = (Widget) NULL;

    // Create 3d Widget.
    arg_cnt = 0;
    XtSetArg(threeD_args[arg_cnt], XtNleft, XtChainLeft); arg_cnt++;
    XtSetArg(threeD_args[arg_cnt], XtNright, XtChainRight); arg_cnt++;
    XtSetArg(threeD_args[arg_cnt], XtNtop, XtChainTop); arg_cnt++;
    XtSetArg(threeD_args[arg_cnt], XtNbottom, XtChainBottom); arg_cnt++;
    XtSetArg(threeD_args[arg_cnt], XtNwidth, w->core.width); arg_cnt++;
    XtSetArg(threeD_args[arg_cnt], XtNheight, w->core.height); arg_cnt++;
    XtSetArg(threeD_args[arg_cnt], XtNrelief, XtReliefSunken); arg_cnt++;
    w->xkviewport.threeD = (ThreeDWidget) XtCreateManagedWidget("threeD",
		threeDWidgetClass,
		new,
		threeD_args, arg_cnt);
    XtVaGetValues((Widget) w->xkviewport.threeD, XtNshadowWidth, &sw, NULL);

    // Create Clip Widget.
    arg_cnt = 0;
    XtSetArg(clip_args[arg_cnt], XtNbackgroundPixmap, None); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNborderWidth, 0); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNleft, XtChainLeft + sw); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNright, XtChainRight - sw); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNtop, XtChainTop + sw); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNbottom, XtChainBottom - sw); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNwidth, w->core.width - 2 * sw); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNheight, w->core.height - 2 * sw); arg_cnt++;
    w->xkviewport.clip = XtCreateManagedWidget("clip",
		widgetClass,
		new,
		clip_args, arg_cnt);

	// If we are not forcing the bars then we are done.
    if (!(w->xkviewport.forcebars)) { return; }

    if ((w->xkviewport.allowhoriz)) { (void) CreateScrollbar(w, True); }
    if ((w->xkviewport.allowvert)) { (void) CreateScrollbar(w, False); }

    h_bar = w->xkviewport.horiz_bar;
    v_bar = w->xkviewport.vert_bar;

	// Set the clip widget to the correct height.
    clip_width = w->core.width - 2 * sw;
    clip_height = w->core.height - 2 * sw;

    if ((h_bar) &&
		((int) w->core.width >
			(int) (h_bar->core.width + h_bar->core.border_width)))
        { clip_width -= h_bar->core.width + h_bar->core.border_width; }
    
    if ((v_bar) &&
		((int) w->core.height >
			(int) (v_bar->core.height + v_bar->core.border_width)))
        { clip_height -= v_bar->core.height + v_bar->core.border_width; }

    arg_cnt = 0;
    XtSetArg(clip_args[arg_cnt], XtNwidth, clip_width); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNheight, clip_height); arg_cnt++;
    XtSetValues(w->xkviewport.clip, clip_args, arg_cnt);
}

/* ARGSUSED */
static void ConstraintInitialize(request, new, args, num_args)
    Widget request, new;
    ArgList args;
    Cardinal *num_args;
{
    ((XKViewportConstraints) new->core.constraints)->xkviewport.reparented =
		False;
}

static void Realize(widget, value_mask, attributes)
    Widget widget;
    XtValueMask *value_mask;
    XSetWindowAttributes *attributes;
{
    XKViewportWidget w = (XKViewportWidget)widget;
    Widget child = w->xkviewport.child;
    Widget clip = w->xkviewport.clip;
    Widget threeD = (Widget)w->xkviewport.threeD;

    *value_mask |= CWBitGravity;
    attributes->bit_gravity = NorthWestGravity;
    (*superclass->core_class.realize)(widget, value_mask, attributes);

    (*w->core.widget_class->core_class.resize)(widget);	/* turn on bars */

    XtRealizeWidget(threeD);
    XLowerWindow(XtDisplay(threeD), XtWindow(threeD));
    if (child != (Widget) NULL)
	{
		XtMoveWidget(child, (Position) 0, (Position) 0);
		XtRealizeWidget(clip);
		XtRealizeWidget(child);
		XReparentWindow(XtDisplay(w), XtWindow(child), XtWindow(clip),
			(Position) 0, (Position) 0);
		XtMapWidget(child);
    }
}

/* ARGSUSED */
static Boolean SetValues(current, request, new, 
args, num_args)
    Widget current, request, new;
    ArgList args;
    Cardinal *num_args;
{
    XKViewportWidget	w = (XKViewportWidget) new;
    XKViewportWidget	cw = (XKViewportWidget) current;

	// Recompute layout.
    if ((w->xkviewport.forcebars != cw->xkviewport.forcebars) ||
			(w->xkviewport.allowvert != cw->xkviewport.allowvert) ||
			(w->xkviewport.allowhoriz != cw->xkviewport.allowhoriz) ||
			(w->xkviewport.useright != cw->xkviewport.useright) ||
			(w->xkviewport.usebottom != cw->xkviewport.usebottom))
		{ (*w->core.widget_class->core_class.resize)(new); }
    return False;
}

static void ChangeManaged(widget)
    Widget widget;
{
    XKViewportWidget	w = (XKViewportWidget)widget;
    int					num_children = w->composite.num_children;
    Widget				child, *childP;
    int					i;
    Dimension			sw;

    child = (Widget) NULL;
    for (childP = w->composite.children, i = 0;
		i < num_children;
		childP++, i++)
	{
		if (XtIsManaged(*childP) &&
			*childP != w->xkviewport.clip &&
			*childP != w->xkviewport.horiz_bar &&
			*childP != w->xkviewport.vert_bar &&
			*childP != (Widget) w->xkviewport.threeD)
			{ child = *childP; break; }
    }

    /*
	XtVaGetValues((Widget)(w->xkviewport.threeD),XtNshadowWidth, &sw,NULL);
    if (XtIsRealized(w->xkviewport.threeD))
	{
        XReparentWindow(XtDisplay(w), XtWindow(w->xkviewport.clip),
			XtWindow(w->xkviewport.threeD),
			(Position) sw, (Position) sw);
    }
	*/

    if (child != w->xkviewport.child) {
	w->xkviewport.child = child;
	if ((child))
	{
	    XtResizeWidget(child, child->core.width, child->core.height,
			(Dimension) 0);
	    if (XtIsRealized(widget))
		{
			XKViewportConstraints constraints =
				(XKViewportConstraints)child->core.constraints;
			if (!XtIsRealized(child))
			{
				Window window = XtWindow(w);
				XtMoveWidget(child, (Position) 0, (Position) 0);
#ifdef notdef
				// this is dirty, but it saves the following code:
				XtRealizeWidget(child);
				XReparentWindow(XtDisplay(w), XtWindow(child),
				     XtWindow(w->xkviewport.clip),
				     (Position)0, (Position)0 );
				if (child->core.mapped_when_managed) { XtMapWidget(child); }
#else 
				w->core.window = XtWindow(w->xkviewport.clip);
				XtRealizeWidget(child);
				w->core.window = window;
#endif /* notdef */
				constraints->xkviewport.reparented = True;
			}
			else if (!constraints->xkviewport.reparented)
			{
				XReparentWindow(XtDisplay(w), XtWindow(child),
				     XtWindow(w->xkviewport.clip),
				     (Position) 0, (Position) 0);
				constraints->xkviewport.reparented = True;
				if (child->core.mapped_when_managed) { XtMapWidget(child); }
			}
	    }
	    GetGeometry(widget, child->core.width, child->core.height);
	    (*((XKViewportWidgetClass)w->core.widget_class)->form_class.layout)
			((FormWidget) w, w->core.width, w->core.height);
		// %%% do we need to hide this child from Form?
	}
    }

#ifdef notdef
    (*superclass->composite_class.change_managed)(widget);
#endif
}

static void SetBar(Widget bar, XKViewportWidget w, Position top,
	Dimension length, Dimension total)
{
	Dimension	sw;

    XtVaGetValues((Widget) w->xkviewport.threeD,
		XtNshadowWidth, (XtArgVal) &sw, NULL);

	length += 2 * sw;
    XKScrollbarSetThumb(bar, (float) top / (float) total,
		 (float) length / (float) total);
}

static void RedrawThumbs(w)
  XKViewportWidget w;
{
    Widget child = w->xkviewport.child;
    Widget clip = w->xkviewport.clip;

    if ((w->xkviewport.horiz_bar))
	{
		SetBar(w->xkviewport.horiz_bar, w, -(child->core.x), clip->core.width,
			child->core.width);
	}

    if ((w->xkviewport.vert_bar))
	{
		SetBar(w->xkviewport.vert_bar, w, -(child->core.y), clip->core.height,
		// ####
		// SetBar(w->xkviewport.vert_bar, w, -(child->core.y), w->core.height,
			child->core.height);
	}
}

static void SendReport(w, changed)
    XKViewportWidget w;
    unsigned int changed;
{
    XawPannerReport rep;

    if ((w->xkviewport.report_callbacks))
	{
		Widget child = w->xkviewport.child;
		Widget clip = w->xkviewport.clip;

		rep.changed = changed;
		rep.slider_x = -child->core.x;	/* child is canvas */
		rep.slider_y = -child->core.y;	/* clip is slider */
		rep.slider_width = clip->core.width;
		rep.slider_height = clip->core.height;
		rep.canvas_width = child->core.width;
		rep.canvas_height = child->core.height;
		XtCallCallbackList ((Widget) w, w->xkviewport.report_callbacks,
			(XtPointer) &rep);
    }
}

static void MoveChild(w, x, y)
    XKViewportWidget w;
    Position x, y;
{
    Widget child = w->xkviewport.child;
    Widget clip = w->xkviewport.clip;

    /* make sure we never move past right/bottom borders */
    if (-x + (int)clip->core.width > (int)child->core.width)
		{ x = -(child->core.width - clip->core.width); }

    if (-y + (int)clip->core.height > (int)child->core.height)
		{ y = -(child->core.height - clip->core.height); }

    /* make sure we never move past left/top borders */
    if (x >= 0) { x = 0; }
    if (y >= 0) { y = 0; }

    XtMoveWidget(child, x, y);
    SendReport(w, (XawPRSliderX | XawPRSliderY));

    RedrawThumbs(w);
}

static void ComputeLayout(widget, query, destroy_scrollbars)
    Widget widget;		/* Viewport */
    Boolean query;		/* query child's preferred geom? */
    Boolean destroy_scrollbars;	/* destroy un-needed scrollbars? */
{
    XKViewportWidget		w = (XKViewportWidget)widget;
    Widget					child = w->xkviewport.child;
    Widget					clip = w->xkviewport.clip;
    Widget					threeD = (Widget) w->xkviewport.threeD;
    XKViewportConstraints	constraints
		= (XKViewportConstraints) clip->core.constraints;
    Boolean					needshoriz, needsvert;
    int						clip_width, clip_height;
    XtWidgetGeometry		intended;
    Dimension				sw = 0;

    XtVaGetValues((Widget) w->xkviewport.threeD, XtNshadowWidth, &sw, NULL);

    if (!(child)) { return; }

    clip_width = w->core.width - 2 * sw;
    clip_height = w->core.height - 2 * sw;
    intended.request_mode = CWBorderWidth;
    intended.border_width = 0;

    if ((w->xkviewport.forcebars))
	{
        needsvert = w->xkviewport.allowvert;
        needshoriz = w->xkviewport.allowhoriz;
        ComputeWithForceBars(widget, query, &intended,
			&clip_width, &clip_height);
    }
    else {
        Dimension prev_width, prev_height;
	XtGeometryMask prev_mode;
	XtWidgetGeometry preferred;

	needshoriz = needsvert = False;

	// Intended.{width,height} caches the eventual child dimensions,
	// but we don't set the mode bits until after we decide that the
	// child's preferences are not acceptable.
	if (!w->xkviewport.allowhoriz) 
	    { intended.request_mode |= CWWidth; }

	if ((int)child->core.width < clip_width) 
	    { intended.width = clip_width; }
	else
	    { intended.width = child->core.width; }

	if ((int)child->core.height < clip_height) 
	    { intended.height = clip_height; }
	else
	    { intended.height = child->core.height; }

	if (!w->xkviewport.allowvert) 
	    { intended.request_mode |= CWHeight; }

	if (!(query))
	{
	    preferred.width = child->core.width;
	    preferred.height = child->core.height;
	}
	do { /* while intended != prev  */

	    if (query) {
	        (void) XtQueryGeometry( child, &intended, &preferred );
		if ( !(preferred.request_mode & CWWidth) )
		    preferred.width = intended.width;
		if ( !(preferred.request_mode & CWHeight) )
		    preferred.height = intended.height;
	    }
	    prev_width = intended.width;
	    prev_height = intended.height;
	    prev_mode = intended.request_mode;
	    /*
	     * note that having once decided to turn on either bar
	     * we'll not change our mind until we're next resized,
	     * thus avoiding potential oscillations.
	     */
#define CheckHoriz()							  \
	    if ( w->xkviewport.allowhoriz &&				  \
		 (int)preferred.width > clip_width) { 			  \
	        if (!needshoriz) {					  \
		    Widget bar;					          \
		    needshoriz = True;				          \
		    if ((bar = w->xkviewport.horiz_bar) == (Widget)NULL)    \
		        bar = CreateScrollbar(w, True);		          \
		    clip_height -= bar->core.height +		          \
			           bar->core.border_width;		  \
		    if (clip_height < 1) clip_height = 1;		  \
		}							  \
	        intended.width = preferred.width;			  \
	    }
/*enddef*/
	    CheckHoriz();
	    if (w->xkviewport.allowvert && (int) preferred.height > clip_height) {
	        if (!needsvert) {
		    Widget bar;
		    needsvert = True;
		    if ((bar = w->xkviewport.vert_bar) == (Widget)NULL)
		        bar = CreateScrollbar(w, False);
		    clip_width -= bar->core.width + bar->core.border_width;
		    if (clip_width < 1) clip_width = 1;
		    CheckHoriz();
		}
		intended.height = preferred.height;
	    }
	    if ( !w->xkviewport.allowhoriz ||
		 (int)preferred.width < clip_width) {
	        intended.width = clip_width;
		intended.request_mode |= CWWidth;
	    }
	    if ( !w->xkviewport.allowvert ||
		 (int)preferred.height < clip_height) {
	        intended.height = clip_height;
		intended.request_mode |= CWHeight;
	    }
	} while (intended.request_mode != prev_mode
		 || (intended.request_mode & CWWidth
		     && intended.width != prev_width)
		 || (intended.request_mode & CWHeight
		     && intended.height != prev_height));
    }

    if (XtIsRealized(threeD))
		{ XLowerWindow(XtDisplay(threeD), XtWindow(threeD)); }

    XtMoveWidget(threeD, (Position) 0, (Position) 0);
    XtResizeWidget(threeD, w->core.width, w->core.height, 0);

    if (XtIsRealized(clip))
		{ XRaiseWindow(XtDisplay(clip), XtWindow(clip)); }

    XtMoveWidget(clip,
		  (Position) (needsvert ? (w->xkviewport.useright ? sw :
			       w->xkviewport.vert_bar->core.width +
			       w->xkviewport.vert_bar->core.border_width+sw) : sw),
		  (Position) (needshoriz ? (w->xkviewport.usebottom ? sw :
				w->xkviewport.horiz_bar->core.height +
			        w->xkviewport.horiz_bar->core.border_width+sw) : sw));
    XtResizeWidget( clip, (Dimension)clip_width,
		    (Dimension)clip_height, (Dimension)0 );
	
    if (w->xkviewport.horiz_bar != (Widget)NULL) {
	Widget bar = w->xkviewport.horiz_bar;
	if (!needshoriz) {
	    constraints->form.vert_base = (Widget)NULL;
	    if (destroy_scrollbars) {
		XtDestroyWidget( bar );
		w->xkviewport.horiz_bar = (Widget)NULL;
	    }
	}
	else {
	    int bw = bar->core.border_width;
	    XtResizeWidget( bar, (Dimension) clip_width, bar->core.height, (Dimension) bw );
	    XtMoveWidget( bar,
			  (Position)((needsvert && !w->xkviewport.useright)
			   ? w->xkviewport.vert_bar->core.width+sw
			   : -bw+sw),
			  (Position)(w->xkviewport.usebottom
			    ? w->core.height - bar->core.height - bw-sw
			    : -bw+sw) );
	    XtSetMappedWhenManaged( bar, True );
	}
    }

    if ((w->xkviewport.vert_bar))
	{
		Widget	bar = w->xkviewport.vert_bar;

		if (!(needsvert))
		{
			constraints->form.horiz_base = (Widget) NULL;
			if (destroy_scrollbars)
			{
				XtDestroyWidget(bar);
				w->xkviewport.vert_bar = (Widget) NULL;
			}
		}
		else
		{
			int		bw = bar->core.border_width;

			// XtResizeWidget(bar, bar->core.width, (Dimension)clip_height, (Dimension)bw );
			XtResizeWidget(bar, bar->core.width,
				(Dimension) w->core.height - 2 * sw, (Dimension) bw);

			XtMoveWidget(bar,
			  (Position)(w->xkviewport.useright
			   ? w->core.width - bar->core.width - bw -sw
			   : -bw+sw),
			  (Position)((needshoriz && !w->xkviewport.usebottom)
			    ? w->xkviewport.horiz_bar->core.height+sw
			    : -bw+sw) );
			XtSetMappedWhenManaged( bar, True );
		}
    }

    if ((child))
	{
		XtResizeWidget(child, (Dimension) intended.width,
			(Dimension)intended.height, (Dimension) 0);
		MoveChild(w,
		  needshoriz ? child->core.x : 0,
		  needsvert ? child->core.y : 0);
    }

    SendReport(w, XawPRAll);
}

/*      Function Name: ComputeWithForceBars
 *      Description: Computes the layout give forcebars is set.
 *      Arguments: widget - the xkviewport widget.
 *                 query - whether or not to query the child.
 *                 intended - the cache of the childs height is
 *                            stored here ( USED AND RETURNED ).
 *                 clip_width, clip_height - size of clip window.
 *                                           (USED AND RETURNED ).
 *      Returns: none.
 */

static void
ComputeWithForceBars(widget, query, intended, clip_width, clip_height)
Widget widget;
Boolean query;
XtWidgetGeometry * intended;
int *clip_width, *clip_height;
{
    XKViewportWidget w = (XKViewportWidget)widget;
    Widget child = w->xkviewport.child;
    XtWidgetGeometry preferred;

	// If forcebars then needs = allows = has.
	// Thus if needsvert is set it MUST have a scrollbar.

    if (w->xkviewport.allowvert) {
	if (w->xkviewport.vert_bar == NULL) 
	    w->xkviewport.vert_bar = CreateScrollbar(w, False);

	*clip_width -= w->xkviewport.vert_bar->core.width +
		       w->xkviewport.vert_bar->core.border_width;
    }

    if (w->xkviewport.allowhoriz) {
	if (w->xkviewport.horiz_bar == NULL) 
	    w->xkviewport.horiz_bar = CreateScrollbar(w, True);

        *clip_height -= w->xkviewport.horiz_bar->core.height +
		       w->xkviewport.horiz_bar->core.border_width;
    }

    AssignMax(*clip_width, 1);
    AssignMax(*clip_height, 1);

    if (!w->xkviewport.allowvert) {
        intended->height = *clip_height;
        intended->request_mode = CWHeight;
    }
    if (!w->xkviewport.allowhoriz) {
        intended->width = *clip_width;
        intended->request_mode = CWWidth;
    }

    if ( query ) {
        if ( (w->xkviewport.allowvert || w->xkviewport.allowhoriz) ) { 
	    XtQueryGeometry( child, intended, &preferred );
	  
	    if ( !(intended->request_mode & CWWidth) )
	        if ( preferred.request_mode & CWWidth )
		    intended->width = preferred.width;
		else
		    intended->width = child->core.width;

	    if ( !(intended->request_mode & CWHeight) )
	        if ( preferred.request_mode & CWHeight )
		    intended->height = preferred.height;
		else
		    intended->height = child->core.height;
	}
    }
    else {
        if (w->xkviewport.allowvert)
	    intended->height = child->core.height;
	if (w->xkviewport.allowhoriz)
	    intended->width = child->core.width;
    }

    if (*clip_width > (int)intended->width)
	intended->width = *clip_width;
    if (*clip_height > (int)intended->height)
	intended->height = *clip_height;
}

static void Resize(widget)
    Widget widget;
	{ ComputeLayout(widget, /*query=*/True, /*destroy=*/True); }


/* ARGSUSED */
static Boolean Layout(w, width, height)
    FormWidget w;
    Dimension width, height;
{
    ComputeLayout((Widget) w, /*query=*/True, /*destroy=*/True);
    w->form.preferred_width = w->core.width;
    w->form.preferred_height = w->core.height;
    return False;
}

static void ScrollUpDownProc(widget, closure, call_data)
    Widget widget;
    XtPointer closure;
    XtPointer call_data;
{
    XKViewportWidget	w = (XKViewportWidget) closure;
    Widget				child = w->xkviewport.child;
    int					pix = (int) call_data;
    Position			x, y;

    if (child == NULL) return;	/* no child to scroll. */

    x = child->core.x - ((widget == w->xkviewport.horiz_bar) ? pix : 0);
    y = child->core.y - ((widget == w->xkviewport.vert_bar) ? pix : 0);
    MoveChild(w, x, y);
}

/* ARGSUSED */
static void ThumbProc(widget, closure, call_data)
    Widget widget;
    XtPointer closure;
    XtPointer call_data;
{
    XKViewportWidget w = (XKViewportWidget)closure;
    Widget child = w->xkviewport.child;
    float *percent = (float *) call_data;
    Position x, y;

    if (child == NULL) return;	/* no child to scroll. */

    if (widget == w->xkviewport.horiz_bar)
#ifdef macII				/* bug in the macII A/UX 1.0 cc */
	x = (int)(-*percent * child->core.width);
#else /* else not macII */
	x = -(int)(*percent * child->core.width);
#endif /* macII */
    else
	x = child->core.x;

    if (widget == w->xkviewport.vert_bar)
#ifdef macII				/* bug in the macII A/UX 1.0 cc */
	y = (int)(-*percent * child->core.height);
#else /* else not macII */
	y = -(int)(*percent * child->core.height);
#endif /* macII */
    else
	y = child->core.y;

    MoveChild(w, x, y);
}

static XtGeometryResult
TestSmaller(w, request, reply_return)
     XKViewportWidget w; XtWidgetGeometry *request, *reply_return;
{
  if (request->width < w->core.width || request->height < w->core.height)
    return XtMakeGeometryRequest((Widget)w, request, reply_return);
  else
    return XtGeometryYes;  
}

static XtGeometryResult
GeometryRequestPlusScrollbar(w, horizontal, request, reply_return)
     Boolean horizontal;
     XKViewportWidget w; 
     XtWidgetGeometry *request, *reply_return;
{
  Widget sb;
  XtWidgetGeometry plusScrollbars;
  plusScrollbars = *request;
  if ((sb = w->xkviewport.horiz_bar) == (Widget)NULL)
    sb = CreateScrollbar( w, horizontal);
  request->width += sb->core.width;
  request->height += sb->core.height;
  XtDestroyWidget(sb);
  return XtMakeGeometryRequest((Widget) w, &plusScrollbars, reply_return);
 }

#define WidthChange() (request->width != w->core.width)
#define HeightChange() (request->height != w->core.height)

static XtGeometryResult 
QueryGeometry(w, request, reply_return)
     XKViewportWidget w; XtWidgetGeometry *request, *reply_return;
{	
  if (w->xkviewport.allowhoriz && w->xkviewport.allowvert) 
    return TestSmaller(w, request, reply_return);

  else if (w->xkviewport.allowhoriz && !w->xkviewport.allowvert) {
    if (WidthChange() && !HeightChange())
      return TestSmaller(w, request, reply_return);
    else if (!WidthChange() && HeightChange())
      return XtMakeGeometryRequest((Widget) w, request, reply_return);
    else if (WidthChange() && HeightChange()) /* hard part */
      return GeometryRequestPlusScrollbar(w, True, request, reply_return);
    else /* !WidthChange() && !HeightChange() */
      return XtGeometryYes;
  }
  else if (!w->xkviewport.allowhoriz && w->xkviewport.allowvert) {
    if (!WidthChange() && HeightChange())
      return TestSmaller(w, request, reply_return);
    else if (WidthChange() && !HeightChange())
      return XtMakeGeometryRequest((Widget)w, request, reply_return);
    else if (WidthChange() && HeightChange()) /* hard part */
      return GeometryRequestPlusScrollbar(w, False, request, reply_return);
    else /* !WidthChange() && !HeightChange() */
      return XtGeometryYes;
  }      
  else /* (!w->xkviewport.allowhoriz && !w->xkviewport.allowvert) */
    return XtMakeGeometryRequest((Widget) w, request, reply_return);
}

#undef WidthChange
#undef HeightChange

static XtGeometryResult GeometryManager(child, request, reply)
    Widget child;
    XtWidgetGeometry *request, *reply;
{
    XKViewportWidget w = (XKViewportWidget)child->core.parent;
    Boolean rWidth = (Boolean)(request->request_mode & CWWidth);
    Boolean rHeight = (Boolean)(request->request_mode & CWHeight);
    XtWidgetGeometry allowed;
    XtGeometryResult result;
    Boolean reconfigured;
    Boolean child_changed_size;
    Dimension height_remaining;

    if (request->request_mode & XtCWQueryOnly)
      return QueryGeometry(w, request, reply);

    if (child != w->xkviewport.child
        || request->request_mode & ~(CWWidth | CWHeight
				     | CWBorderWidth)
	|| ((request->request_mode & CWBorderWidth)
	    && request->border_width > 0))
	return XtGeometryNo;

    allowed = *request;

    reconfigured = GetGeometry( (Widget)w,
			        (rWidth ? request->width : w->core.width),
			        (rHeight ? request->height : w->core.height)
			      );

    child_changed_size = ((rWidth && child->core.width != request->width) ||
			  (rHeight && child->core.height != request->height));

    height_remaining = w->core.height;
    if (rWidth && w->core.width != request->width) {
	if (w->xkviewport.allowhoriz && request->width > w->core.width) {
	    /* horizontal scrollbar will be needed so possibly reduce height */
	    Widget bar; 
	    if ((bar = w->xkviewport.horiz_bar) == (Widget)NULL)
		bar = CreateScrollbar( w, True );
	    height_remaining -= bar->core.height + bar->core.border_width;
	    reconfigured = True;
	}
	else {
	    allowed.width = w->core.width;
	}
    }
    if (rHeight && height_remaining != request->height) {
	if (w->xkviewport.allowvert && request->height > height_remaining) {
	    /* vertical scrollbar will be needed, so possibly reduce width */
	    if (!w->xkviewport.allowhoriz || request->width < w->core.width) {
		Widget bar;
		if ((bar = w->xkviewport.vert_bar) == (Widget)NULL)
		    bar = CreateScrollbar( w, False );
		if (!rWidth) {
		    allowed.width = w->core.width;
		    allowed.request_mode |= CWWidth;
		}
		if ( (int)allowed.width >
		     (int)(bar->core.width + bar->core.border_width) )
		    allowed.width -= bar->core.width + bar->core.border_width;
		else
		    allowed.width = 1;
		reconfigured = True;
	    }
	}
	else {
	    allowed.height = height_remaining;
	}
    }

    if (allowed.width != request->width || allowed.height != request->height) {
	*reply = allowed;
	result = XtGeometryAlmost;
    }
    else {
	if (rWidth)  child->core.width = request->width;
	if (rHeight) child->core.height = request->height;
	result = XtGeometryYes;
    }

    if (reconfigured || child_changed_size)
	ComputeLayout( (Widget)w,
		       /*query=*/ False,
		       /*destroy=*/ (result == XtGeometryYes) ? True : False );

    return result;
}

static Boolean GetGeometry(w, width, height)
    Widget w;
    Dimension width, height;
{
    XtWidgetGeometry geometry, return_geom;
    XtGeometryResult result;

    if (width == w->core.width && height == w->core.height)
	return False;

    geometry.request_mode = CWWidth | CWHeight;
    geometry.width = width;
    geometry.height = height;

    if (XtIsRealized(w))
	{
		if (((XKViewportWidget) w)->xkviewport.allowhoriz &&
				width > w->core.width)
			{ geometry.width = w->core.width; }
		if (((XKViewportWidget) w)->xkviewport.allowvert &&
				height > w->core.height)
			{ geometry.height = w->core.height; }
    }
	else
	{
		// This is the Realize call; we'll inherit a w&h iff none currently.
		if (w->core.width != 0)
		{
			if (w->core.height != 0) { return False; }
			geometry.width = w->core.width;
		}
		if (w->core.height != 0) { geometry.height = w->core.height; }
    }

    result = XtMakeGeometryRequest(w, &geometry, &return_geom);
    if (result == XtGeometryAlmost)
	{ 
		result =
			XtMakeGeometryRequest(w, &return_geom, (XtWidgetGeometry *) NULL);
	}
    return (result == XtGeometryYes);
}

static XtGeometryResult PreferredGeometry(w, constraints, reply)
    Widget w;
    XtWidgetGeometry *constraints, *reply;
{
    if (((XKViewportWidget) w)->xkviewport.child != NULL)
	{
		return XtQueryGeometry(((XKViewportWidget) w)->xkviewport.child,
			constraints, reply );
	} 
    else
		{ return XtGeometryYes; }
}


void XawViewportSetLocation (Widget gw, float xoff, float yoff)
{
    XKViewportWidget w = (XKViewportWidget) gw;
    Widget child = w->xkviewport.child;
    Position x, y;

    if (xoff > 1.0)			/* scroll to right */
       { x = child->core.width; }
    else if (xoff < 0.0)		/* if the offset is < 0.0 nothing */ 
       { x = child->core.x; }
    else
       { x = (Position) (((float) child->core.width) * xoff); }
    if (yoff > 1.0) 
       { y = child->core.height; }
    else if (yoff < 0.0)
       { y = child->core.y; }
    else
       { y = (Position) (((float) child->core.height) * yoff); }
    MoveChild (w, -x, -y);
}

void XKViewportSetCoordinates(Widget gw, Position x, Position y)
{
    XKViewportWidget w = (XKViewportWidget) gw;
    Widget child = w->xkviewport.child;

    if (x > (int)child->core.width) 
		{ x = child->core.width; }
    else if (x < 0)
		{ x = child->core.x; }
    if (y > (int)child->core.height)
		{ y = child->core.height; }
    else if (y < 0)
		{ y = child->core.y; }
    MoveChild (w, -x, -y);
}

